"""Module containing functionality for the deprecated INI-based configuration format."""

import sys
from configparser import ConfigParser, ExtendedInterpolation, SectionProxy
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from rust2rpm import log
from rust2rpm.sysinfo import TARGET_NAMES

from .common import ConfError
from .toml import TomlConf

VALID_INI_KEYS = [
    "summary",
    "supported-arches",
    "all-features",
    "unwanted-features",
    "enabled-features",
    "buildrequires",
    "testrequires",
    "lib.requires",
    "bin.requires",
]


def to_list(s: str, *, sort: bool = True) -> list[str]:
    """Parse a multi-line value into a list of strings."""
    f = sorted if sort else list
    return f(filter(None, (elem.strip() for elem in s.splitlines())))


def _validate_ini_conf_with_features(merged: SectionProxy, validate_features: set[str]):
    # validate configuration file
    valid_keys = VALID_INI_KEYS.copy()

    # the "default" feature is always implicitly defined
    valid_keys.append("lib+default.requires")
    valid_keys.extend(f"lib+{feature}.requires" for feature in validate_features)

    # check setting keys
    for key in merged:
        if key not in valid_keys:
            msg = f"Invalid key: {key!r}"
            raise ConfError(msg)

    # check settings values
    if unwanted_features := merged.get("unwanted-features"):
        unwanted_features = to_list(unwanted_features)
        for unwanted_feature in unwanted_features:
            if unwanted_feature not in validate_features:
                msg = f'Unrecognized "unwanted" feature: {unwanted_feature!r}'
                raise ConfError(msg)

    if enabled_features := merged.get("enabled-features"):
        enabled_features = to_list(enabled_features)
        for enabled_feature in enabled_features:
            if enabled_feature not in validate_features:
                msg = f'Unrecognized "enabled" feature: {enabled_feature!r}'
                raise ConfError(msg)


def _validate_ini_conf_without_features(merged: SectionProxy) -> set[str]:
    # validate configuration file
    valid_keys = VALID_INI_KEYS.copy()

    # determine features that are mentioned in the configuration file
    mentioned_features = set()
    for key in merged:
        if key.startswith("lib+") and key.endswith(".requires"):
            mentioned_features.add(key.removeprefix("lib+").removesuffix(".requires"))

    # check setting keys
    for key in merged:
        # cannot validate against actual crate features:
        # just validate against *mentioned* features and check for valid syntax
        if key not in valid_keys:
            if key in [f"lib+{feature}.requires" for feature in mentioned_features]:
                continue
            msg = f"Invalid key: {key!r}"
            raise ConfError(msg)

    return mentioned_features


@dataclass(frozen=True)
class IniConf:
    """rust2rpm configuration loaded from the deprecated rust2rpm.conf file format."""

    summary: str | None = None
    """Override for the generated RPM Summary tag."""

    supported_arches: list[str] | None = None
    """List of architectures supported by the package.

    For crates, this is used to conditionally run `%build` and `%check` scriptlets.
    For other projects, it is used as the value of an `ExclusiveArch` tag.
    """

    all_features: bool | None = None
    """Flag that controls whether the `--all-features` flag is passed to cargo."""

    unwanted_features: list[str] | None = None
    """List of features for which the generation of the corresponding subpackage is suppressed."""

    enabled_features: list[str] | None = None
    """List of features that are enabled and passed with `--features` to cargo."""

    buildrequires: list[str] | None = None
    """List of additional `BuildRequires`."""

    testrequires: list[str] | None = None
    """List of additional `BuildRequires` that only apply when building the package with tests."""

    bin_requires: list[str] | None = None
    """List of additional `Requires` for the built binary subpackage."""

    lib_requires: dict[str | None, list[str]] | None = None
    """List of additional `Requires` for built library subpackages."""

    @staticmethod
    def _from_configparser(merged: SectionProxy, validate_features: set[str] | None) -> "IniConf":
        # validate configuration
        if validate_features is not None:
            _validate_ini_conf_with_features(merged, validate_features)
            known_features = validate_features
        else:
            mentioned_features = _validate_ini_conf_without_features(merged)
            known_features = mentioned_features

        # parse configuration
        summary = value if (value := merged.get("summary")) else None
        supported_arches = to_list(value, sort=False) if (value := merged.get("supported-arches")) else None

        all_features = value if (value := merged.getboolean("all-features")) and value is not None else None
        unwanted_features = to_list(value) if (value := merged.get("unwanted-features")) else None
        enabled_features = to_list(value) if (value := merged.get("enabled-features")) else None

        buildrequires = to_list(value) if (value := merged.get("buildrequires")) else None
        testrequires = to_list(value) if (value := merged.get("testrequires")) else None
        bin_requires = to_list(value) if (value := merged.get("bin.requires")) else None

        lib_requires: dict[str | None, list[str]] | None = None
        if value := merged.get("lib.requires"):
            lib_requires = {}
            lib_requires[None] = to_list(value)  # type: ignore[index]

        for feature in known_features:
            if value := merged.get(f"lib+{feature}.requires"):
                if lib_requires is None:
                    lib_requires = {}
                lib_requires[feature] = to_list(value)  # type: ignore[index]

        return IniConf(
            summary,
            supported_arches,
            all_features,
            unwanted_features,
            enabled_features,
            buildrequires,
            testrequires,
            bin_requires,
            lib_requires,
        )

    @staticmethod
    def load(filenames: list[str], target: str, validate_features: set[str] | None) -> "IniConf":
        """Load and validate rust2rpm configuration for the given target.

        Arguments:
            filenames: List of file paths that are attempted to be read.
            target: Name of the spec file generation target.
            validate_features: Optional set of valid feature names.
                Used to validate some settings.

        Returns:
            Loaded and validated rust2rpm configuration.

        Raises:
            `FileNotFoundError` when none of the expected file paths existed.
            `FileExistsError` when more than one config file was found.
            `ConfError` if the configuration file failed validation.

        """
        conf = ConfigParser(interpolation=ExtendedInterpolation())
        confs = conf.read(filenames)

        if len(confs) == 0:
            raise FileNotFoundError

        if len(confs) > 1:  # pragma nocover
            raise FileExistsError

        # clean up configuration files with deprecated names
        if ".rust2rpm.conf" in confs:  # pragma nocover
            Path(".rust2rpm.conf").rename("rust2rpm.conf")
            log.info("Renamed deprecated, hidden .rust2rpm.conf file to rust2rpm.conf.")

        if "_rust2rpm.conf" in confs:  # pragma nocover
            Path("_rust2rpm.conf").rename("rust2rpm.conf")
            log.info("Renamed deprecated _rust2rpm.conf file to rust2rpm.conf.")

        # merge target-specific configuration with default configuration
        if target not in conf:
            conf.add_section(target)
        merged = conf[target]

        # check section names
        for section in conf:
            if section != "DEFAULT" and section not in TARGET_NAMES:
                msg = f"Invalid section: {section!r}"
                raise ConfError(msg)

        return IniConf._from_configparser(merged, validate_features)

    def upgrade(self) -> TomlConf:  # noqa: PLR0912
        """Upgrade from old to new configuration file format."""
        settings: dict[str, Any] = {}

        package: dict[str, Any] = {}
        if self.summary:
            package["summary"] = self.summary
        if self.supported_arches:
            package["supported-arches"] = self.supported_arches
        if package:
            settings["package"] = package

        features: dict[str, Any] = {}
        if self.all_features is not None:
            features["enable-all"] = self.all_features
        if self.enabled_features:
            features["enable"] = self.enabled_features
        if self.unwanted_features:
            features["hide"] = self.unwanted_features
        if features:
            settings["features"] = features

        requires: dict[str, Any] = {}
        if self.buildrequires:
            requires["build"] = self.buildrequires
        if self.testrequires:
            requires["test"] = self.testrequires
        if self.bin_requires:
            requires["bin"] = self.bin_requires
        if self.lib_requires:
            for key, value in self.lib_requires.items():
                if key is None:
                    requires["lib"] = value
                else:
                    if "features" not in requires:
                        requires["features"] = {}
                    requires["features"][key] = value
        if requires:
            settings["requires"] = requires

        # feature names were already validated
        return TomlConf.from_data(settings, None)

    def migrate(self, validate_features: set[str] | None) -> str:
        """Migrate from old to new configuration file format.

        Arguments:
            validate_features: Optional set of valid feature names.
                Used to validate some settings.

        Returns:
            Configuration in rust2rpm.toml format equivalent to the
                settings in the original config file.

        """
        # check if the written config would be valid
        old = self.upgrade()

        lines = []

        def write_array(name: str, values: list[str] | None):
            if values is None:
                return
            lines.append(f"{name} = [")
            lines.extend([f'    "{value}",' for value in values])
            lines.append("]")

        if self.summary or self.supported_arches:
            lines.append("[package]")
            if summary := self.summary:
                lines.append(f'summary = "{summary}"')
            write_array("supported-arches", self.supported_arches)
            lines.append("")

        if self.all_features or self.unwanted_features or self.enabled_features:
            lines.append("[features]")
            if self.all_features is True:
                lines.append("enable-all = true")
            write_array("enable", self.enabled_features)
            write_array("hide", self.unwanted_features)
            lines.append("")

        if self.buildrequires or self.testrequires or self.bin_requires or self.lib_requires:
            lines.append("[requires]")
            write_array("build", self.buildrequires)
            write_array("test", self.testrequires)
            write_array("bin", self.bin_requires)
            if lib_requires := self.lib_requires:
                write_array("lib", lib_requires.pop(None, None))
                for feature, feature_deps in sorted(lib_requires.items()):
                    write_array(f"features.{feature}", feature_deps)
            lines.append("")

        toml = "\n".join(lines)

        new = TomlConf.from_str(toml, validate_features)

        # check if contents are equivalent
        if old != new:  # pragma nocover: no known failure cases
            msg = "Failed to convert rust2rpm.conf to rust2rpm.toml automatically."
            raise ConfError(msg)

        return toml


def load_config_ini(paths: list[str], target: str, validate_features: set[str] | None = None) -> IniConf | None:
    """Load rust2rpm configuration from the deprecated rust2rpm.conf file format.

    Arguments:
        paths: List of accepted file paths (at most one of them should point to
            an existing file).
        target: Spec file generation target.
        validate_features: Optional set of valid feature names.
            Used to validate some settings.

    Returns:
        Configuration loaded from a file in INI format, or `None` if no file was found.

    Raises:
        `SystemExit` on fatal errors.

    """
    try:
        distconf = IniConf.load(paths, target, validate_features)

    except FileNotFoundError:
        return None

    except FileExistsError:
        log.error(
            "More than one *rust2rpm.conf file is present in this directory. "
            "Ensure that there is only one, and that it has the correct contents.",
        )
        sys.exit(1)

    except ConfError as exc:
        log.error("Invalid rust2rpm configuration file:")
        log.error(str(exc))
        sys.exit(1)

    else:
        return distconf
