"""Module containing base parameter classes and functionality for preparing template parameters."""

import time
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Any


class InvalidMetadataError(ValueError):
    """Raised for invalid project metadata (like missing SPDX license expression)."""


class Template(metaclass=ABCMeta):
    """Base class for all templates."""

    @abstractmethod
    def render(self) -> str:
        """Render spec file from template with the computed parameters."""


class Parameters:
    """Base class for all parameter classes."""


def parameters_as_dict(p: Parameters) -> dict[str, Any]:
    """Convert `Parameters` into a dictionary suitable for passing to `jinja2.Template.render`.

    Class attributes that start with underscores are ignored.
    """
    return {attr: getattr(p, attr) for attr in dir(p) if not attr.startswith("_")}


@dataclass(frozen=True)
class TargetParameters(metaclass=ABCMeta):
    """Base class for target-specific parameters."""

    _date: time.struct_time | None
    _packager: str | None
    _use_rpmautospec: bool

    @property
    @abstractmethod
    def rpm_release(self) -> str:
        """RPM package Release (exact value depends on the target)."""

    @property
    @abstractmethod
    def rpm_group(self) -> str | None:
        """RPM package Group (not defined in all targets)."""

    @property
    @abstractmethod
    def include_build_requires(self) -> bool:
        """Toggle between dynamically generated and static BuildRequires."""

    @property
    @abstractmethod
    def include_requires(self) -> bool:
        """Toggle between dynamically generated and static Requires."""

    @property
    @abstractmethod
    def include_provides(self) -> bool:
        """Toggle between dynamically generated and static Provides."""

    @property
    @abstractmethod
    def rpm_changelog_date(self) -> str:
        """Date used in the automatically generated %changelog entry."""

    @property
    @abstractmethod
    def rpm_changelog_packager(self) -> str:
        """Packager identity used in the automatically generated %changelog entry."""

    @property
    @abstractmethod
    def spec_copyright_year(self) -> str | None:
        """Copyright year for templates that include it (opensuse)."""

    @property
    @abstractmethod
    def rpm_buildrequires(self) -> list[str]:
        """Automatically generated RPM BuildRequires for crate dependencies."""

    @property
    @abstractmethod
    def rpm_test_requires(self) -> list[str]:
        """Automatically generated test-only RPM BuildRequires for crate dependencies."""

    @property
    @abstractmethod
    def rpm_requires(self) -> dict[str | None, list[str]]:
        """Automatically generated RPM Requires for subpackages."""

    @property
    @abstractmethod
    def rpm_provides(self) -> dict[str | None, str]:
        """Automatically generated RPM Provides for subpackages."""

    @property
    @abstractmethod
    def use_rpmautospec(self) -> bool:
        """Toggle usage of RPMAutospec in the generated spec file."""


@dataclass(frozen=True)
class ParametersFedora(TargetParameters, metaclass=ABCMeta):
    """Base class for parameters specific to the "fedora" target."""

    @property
    def rpm_release(self) -> str:
        """RPM package Release (exact value depends on the target)."""
        return "%autorelease" if self._use_rpmautospec else "1%{?dist}"

    @property
    def rpm_group(self) -> str | None:
        """RPM package Group (not defined in all targets)."""
        return None

    @property
    def include_build_requires(self) -> bool:
        """Toggle between dynamically generated and static BuildRequires."""
        return False

    @property
    def include_requires(self) -> bool:
        """Toggle between dynamically generated and static Requires."""
        return False

    @property
    def include_provides(self) -> bool:
        """Toggle between dynamically generated and static Provides."""
        return False

    @property
    def rpm_changelog_date(self) -> str:
        """Date used in the automatically generated %changelog entry."""
        date_format = "%a %b %d %Y"
        return time.strftime(date_format, self._date) if self._date else time.strftime(date_format)

    @property
    def rpm_changelog_packager(self) -> str:
        """Packager identity used in the automatically generated %changelog entry."""
        return self._packager or "rust2rpm <nobody@fedoraproject.org>"

    @property
    def spec_copyright_year(self) -> str | None:
        """Copyright year for templates that include it (opensuse)."""
        return None

    @property
    def rpm_buildrequires(self) -> list[str]:
        """Automatically generated RPM BuildRequires for crate dependencies."""
        return []

    @property
    def rpm_test_requires(self) -> list[str]:
        """Automatically generated test-only RPM BuildRequires for crate dependencies."""
        return []

    @property
    def rpm_requires(self) -> dict[str | None, list[str]]:
        """Automatically generated RPM Requires for subpackages."""
        return {}

    @property
    def rpm_provides(self) -> dict[str | None, str]:
        """Automatically generated RPM Provides for subpackages."""
        return {}

    @property
    def use_rpmautospec(self) -> bool:
        """Toggle usage of RPMAutospec in the generated spec file."""
        return self._use_rpmautospec


@dataclass(frozen=True)
class ParametersMageia(TargetParameters, metaclass=ABCMeta):
    """Base class for parameters specific to the "mageia" target."""

    @property
    def rpm_release(self) -> str:
        """RPM package Release (exact value depends on the target)."""
        return "%mkrel 1"

    @property
    def rpm_group(self) -> str | None:
        """RPM package Group (not defined in all targets)."""
        return "Development/Rust"

    @property
    def include_build_requires(self) -> bool:
        """Toggle between dynamically generated and static BuildRequires."""
        return True

    @property
    def include_requires(self) -> bool:
        """Toggle between dynamically generated and static Requires."""
        return False

    @property
    def include_provides(self) -> bool:
        """Toggle between dynamically generated and static Provides."""
        return False

    @property
    def rpm_changelog_date(self) -> str:
        """Date used in the automatically generated %changelog entry."""
        date_format = "%a %b %d %Y"
        return time.strftime(date_format, self._date) if self._date else time.strftime(date_format)

    @property
    def rpm_changelog_packager(self) -> str:
        """Packager identity used in the automatically generated %changelog entry."""
        return self._packager or "rust2rpm <nobody@mageia.org>"

    @property
    def spec_copyright_year(self) -> str | None:
        """Copyright year for templates that include it (opensuse)."""
        return None

    @property
    def rpm_requires(self) -> dict[str | None, list[str]]:
        """Automatically generated RPM Requires for subpackages."""
        return {}

    @property
    def rpm_provides(self) -> dict[str | None, str]:
        """Automatically generated RPM Provides for subpackages."""
        return {}

    @property
    def use_rpmautospec(self) -> bool:
        """Toggle usage of RPMAutospec in the generated spec file."""
        return False


@dataclass(frozen=True)
class ParametersOpenSUSE(TargetParameters, metaclass=ABCMeta):
    """Base class for parameters specific to the "opensuse" target."""

    @property
    def rpm_release(self) -> str:
        """RPM package Release (exact value depends on the target)."""
        return "0"

    @property
    def rpm_group(self) -> str | None:
        """RPM package Group (not defined in all targets)."""
        return "Development/Libraries/Rust"

    @property
    def include_build_requires(self) -> bool:
        """Toggle between dynamically generated and static BuildRequires."""
        return True

    @property
    def include_requires(self) -> bool:
        """Toggle between dynamically generated and static Requires."""
        return False

    @property
    def include_provides(self) -> bool:
        """Toggle between dynamically generated and static Provides."""
        return False

    @property
    def rpm_changelog_date(self) -> str:
        """Date used in the automatically generated %changelog entry."""
        date_format = "%a %b %d %T %Z %Y"
        return time.strftime(date_format, self._date) if self._date else time.strftime(date_format)

    @property
    def rpm_changelog_packager(self) -> str:
        """Packager identity used in the automatically generated %changelog entry."""
        return self._packager or "rust2rpm <opensuse-packaging@opensuse.org>"

    @property
    def spec_copyright_year(self) -> str | None:
        """Copyright year for templates that include it (opensuse)."""
        return time.strftime("%Y")

    @property
    def rpm_requires(self) -> dict[str | None, list[str]]:
        """Automatically generated RPM Requires for subpackages."""
        return {}

    @property
    def rpm_provides(self) -> dict[str | None, str]:
        """Automatically generated RPM Provides for subpackages."""
        return {}

    @property
    def use_rpmautospec(self) -> bool:
        """Toggle usage of RPMAutospec in the generated spec file."""
        return False


@dataclass(frozen=True)
class ParametersEpelEight(TargetParameters, metaclass=ABCMeta):
    """Base class for parameters specific to the "epel8" target."""

    @property
    def rpm_release(self) -> str:
        """RPM package Release (exact value depends on the target)."""
        return "%autorelease" if self._use_rpmautospec else "1%{?dist}"

    @property
    def rpm_group(self) -> str | None:
        """RPM package Group (not defined in all targets)."""
        return None

    @property
    def include_build_requires(self) -> bool:
        """Toggle between dynamically generated and static BuildRequires."""
        return False

    @property
    def include_requires(self) -> bool:
        """Toggle between dynamically generated and static Requires."""
        return False

    @property
    def include_provides(self) -> bool:
        """Toggle between dynamically generated and static Provides."""
        return False

    @property
    def rpm_changelog_date(self) -> str:
        """Date used in the automatically generated %changelog entry."""
        date_format = "%a %b %d %Y"
        return time.strftime(date_format, self._date) if self._date else time.strftime(date_format)

    @property
    def rpm_changelog_packager(self) -> str:
        """Packager identity used in the automatically generated %changelog entry."""
        return self._packager or "rust2rpm <nobody@fedoraproject.org>"

    @property
    def spec_copyright_year(self) -> str | None:
        """Copyright year for templates that include it (opensuse)."""
        return None

    @property
    def rpm_buildrequires(self) -> list[str]:
        """Automatically generated RPM BuildRequires for crate dependencies."""
        return []

    @property
    def rpm_test_requires(self) -> list[str]:
        """Automatically generated test-only RPM BuildRequires for crate dependencies."""
        return []

    @property
    def rpm_requires(self) -> dict[str | None, list[str]]:
        """Automatically generated RPM Requires for subpackages."""
        return {}

    @property
    def rpm_provides(self) -> dict[str | None, str]:
        """Automatically generated RPM Provides for subpackages."""
        return {}

    @property
    def use_rpmautospec(self) -> bool:
        """Toggle usage of RPMAutospec in the generated spec file."""
        return self._use_rpmautospec


@dataclass(frozen=True)
class ParametersPlain(TargetParameters, metaclass=ABCMeta):
    """Base class for parameters specific to the "plain" target."""

    @property
    def rpm_release(self) -> str:
        """RPM package Release (exact value depends on the target)."""
        return "1%{?dist}"

    @property
    def rpm_group(self) -> str | None:
        """RPM package Group (not defined in all targets)."""
        return None

    @property
    def include_build_requires(self) -> bool:
        """Toggle between dynamically generated and static BuildRequires."""
        return True

    @property
    def include_requires(self) -> bool:
        """Toggle between dynamically generated and static Requires."""
        return True

    @property
    def include_provides(self) -> bool:
        """Toggle between dynamically generated and static Provides."""
        return True

    @property
    def rpm_changelog_date(self) -> str:
        """Date used in the automatically generated %changelog entry."""
        date_format = "%a %b %d %Y"
        return time.strftime(date_format, self._date) if self._date else time.strftime(date_format)

    @property
    def rpm_changelog_packager(self) -> str:
        """Packager identity used in the automatically generated %changelog entry."""
        return self._packager or "rust2rpm <packager@example.com>"

    @property
    def spec_copyright_year(self) -> str | None:
        """Copyright year for templates that include it (opensuse)."""
        return time.strftime("%Y")

    @property
    def use_rpmautospec(self) -> bool:
        """Toggle usage of RPMAutospec in the generated spec file."""
        return False
