"""Module containing functionality for patching Cargo.toml files."""

import shutil
import subprocess
import tempfile
import time
from datetime import UTC, datetime
from difflib import unified_diff
from pathlib import Path

from rust2rpm import log
from rust2rpm.cli import Options
from rust2rpm.utils import detect_editor


class PatchError(ValueError):
    """Raised when re-applying an existing patch fails."""


def file_mtime_in_iso_format(path: Path) -> str:
    """Format file modification timestamp in the format expected by patch files."""
    mtime = path.stat().st_mtime
    return datetime.fromtimestamp(mtime, UTC).isoformat()


def make_diff_from_lines(
    diff_path: str,
    lines_before: list[str],
    mtime_before: str,
    lines_after: list[str],
    mtime_after: str,
) -> list[str]:
    """Format a patch file based on the "before" and "after" states.

    Arguments:
        diff_path: Path of the patched file.
        lines_before: File contents (as list of lines) before changes were applied.
        mtime_before: File modification timestamp in the expected format.
        lines_after: File contents (as list of lines) after changes were applied.
        mtime_after: File modification timestamp in the expected format.

    Returns:
        Patch in "unified diff" format (as list of lines).

    """
    return list(
        unified_diff(
            lines_before,
            lines_after,
            fromfile=diff_path,
            tofile=diff_path,
            fromfiledate=mtime_before,
            tofiledate=mtime_after,
            lineterm="",
        ),
    )


def preprocess_cargo_toml(contents: str) -> str | None:
    """Automatically pre-process contents of a Cargo.toml file.

    Arguments:
        contents: Original contents of the Cargo.toml file.

    Returns:
        Modified Cargo.toml contents or `None` if no modifications were made.

    """
    if shutil.which("rust2rpm-helper"):
        return preprocess_cargo_toml_helper(contents)

    log.warn("rust2rpm-helper is not installed. Skipping automatic patching of Cargo.toml.")
    return None


def preprocess_cargo_toml_helper(contents: str) -> str | None:
    """Call rust2rpm-helper executable to pre-process a Cargo.toml file.

    Arguments:
        contents: Original contents of the Cargo.toml file.

    Returns:
        Modified Cargo.toml contents or `None` if no modifications were made.

    """
    # S603: rust2rpm-helper validates input itself
    # S607: allowing use of "rust2rpm-helper" from $PATH is intentional

    ret1 = subprocess.run(  # noqa: S603
        ["rust2rpm-helper", "normalize-version", "-"],  # noqa: S607
        input=contents,
        text=True,
        check=True,
        stdout=subprocess.PIPE,
    )
    patched1 = ret1.stdout or contents

    ret2 = subprocess.run(  # noqa: S603
        ["rust2rpm-helper", "strip-foreign", "-"],  # noqa: S607
        input=patched1,
        text=True,
        check=True,
        stdout=subprocess.PIPE,
    )
    patched2 = ret2.stdout or patched1

    if patched2 == contents:
        return None

    return patched2


def reapply_patch(toml_path: Path, reapply_path: Path) -> bool:
    """Attempt to re-apply an existing patch file.

    Arguments:
        toml_path: Path to the Cargo.toml file.
        reapply_path: Path to the patch file.

    Returns:
        Boolean flag indicating whether the patch applied successfully.

    """
    patch = shutil.which("patch")

    if patch is None:
        msg = "No 'patch' executable found."
        raise PatchError(msg)

    if not reapply_path.exists():
        msg = "No applicable patch file found."
        raise PatchError(msg)

    # copy file to a temporary directory to work around weird behaviour of "patch"
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_toml = Path(temp_dir) / toml_path.name
        shutil.copy2(toml_path, temp_toml)

        try:
            subprocess.run(  # noqa: S603
                [patch, "-p1", str(temp_toml), str(reapply_path)],
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError:
            return False
        else:
            shutil.copy2(temp_toml, toml_path)
            return True


def make_patches(
    name: str,
    version: str,
    toml_path: Path,
    reapply_path: Path | None,
    options: Options,
) -> tuple[list[str] | None, list[str] | None]:
    """Prepare patches for Cargo.toml (automatic and manual changes).

    Any automatically applied changes to Cargo.toml are returned as the first
    patch. Manual changes to Cargo.toml are either based on changes made
    interactively in the opened editor, or based on changes from a successfully
    re-applied existing patch, or both.

    Arguments:
        name: Name of the crate (used to construct the file path in the patch).
        version: Version of the crate (used to construct the file path in the patch).
        toml_path: Path to the Cargo.toml file that will be patched.
        reapply_path: Optional path to an existing patch file that is attempted
            to be reapplied.
        options: Options and flags derived from command-line arguments.

    Returns:
        Tuple of (automatic, manual) patch file contents as lists of lines,
            or `None` in case no changes were made.

    """
    mtime_before = file_mtime_in_iso_format(toml_path)

    with toml_path.open() as file:
        toml_before = file.read()

    diff1 = diff2 = None
    diff_path = f"{name}-{version}/Cargo.toml"

    # patching Cargo.toml involves multiple steps:
    #
    # 1) attempt to automatically drop "foreign" dependencies
    #    If this succeeded, remove the original Cargo.toml file, write the
    #    changed contents back to disk, and generate a diff.
    # 2) if requested, open Cargo.toml in an editor for further, manual changes
    #
    # The changes from *both* steps must be reflected in the Cargo.toml file
    # that ends up on disk after this function returns. Otherwise, the
    # calculated metadata and generated spec file will not reflect the patches
    # to Cargo.toml that were generated here.

    if toml_after := options.patch_foreign and preprocess_cargo_toml(toml_before):
        # remove original Cargo.toml file
        toml_path.unlink()

        # write auto-patched Cargo.toml to disk
        with toml_path.open("w") as file:
            file.write(toml_after)

        mtime_after1 = file_mtime_in_iso_format(toml_path)
        diff1 = make_diff_from_lines(
            diff_path,
            toml_before.splitlines(),
            mtime_before,
            toml_after.splitlines(),
            mtime_after1,
        )
    else:
        toml_after = toml_before

    # early exit
    if not (options.patch or reapply_path):
        return diff1, diff2

    # attempt to reapply existing patch
    if reapply_path and not reapply_patch(toml_path, reapply_path):
        if options.patch:
            log.warn("Re-applying existing patch failed, opening editor instead.")
            # sleep so the user has an actual chance to see this log message
            time.sleep(2)
        else:
            msg = "Failed to re-apply existing patch."
            raise PatchError(msg)

    # open editor for Cargo.toml
    if options.patch:
        editor = detect_editor()
        subprocess.check_call([editor, toml_path])  # noqa: S603

    with toml_path.open() as file:
        toml_after2 = file.read()

    mtime_after2 = file_mtime_in_iso_format(toml_path)
    diff2 = make_diff_from_lines(
        diff_path,
        toml_after.splitlines(),
        mtime_before,
        toml_after2.splitlines(),
        mtime_after2,
    )

    return diff1, diff2
